import { Observable, ReplaySubject, asapScheduler, Subscriber, Subscription } from 'rxjs';
import { delay, filter, tap, map, withLatestFrom, scan } from 'rxjs/operators';
import { keys, map as map$1, flow, bindKey } from 'micro-dash';
import { isSetEqual } from 's-js-utils';
import { __spread, __read } from 'tslib';

/**
 * 1. Caches the last value emitted to give to new subscribers (without running any upstream pipe operators)
 * 2. Manages all subscribers directly, without passing subscriptions up the stream.
 *
 * This is very similar to `shareReplay(1)`, except that once all subscribers unsubscribe this also unsubscribes from the upstream observable.
 *
 * ```ts
 * const source = new BehaviorSubject(1000);
 * const result = source.pipe(expensiveComputation(), cache());
 * source.subscribe(); // expensiveComputation(1000) runs
 * source.subscribe(); // the cached result is used
 * source.next(2000); // expensiveComputation(2000) runs once, emitted to both subscribers
 * ```
 */
function cache() {
    return function (source) {
        var middleMan;
        var upstreamSubscription;
        return new Observable(function (subscriber) {
            if (!middleMan) {
                middleMan = new ReplaySubject(1);
                upstreamSubscription = source.subscribe(middleMan);
            }
            var subscription = middleMan.subscribe(subscriber);
            // teardown logic
            return function () {
                subscription.unsubscribe();
                if (middleMan && middleMan.observers.length === 0) {
                    upstreamSubscription.unsubscribe();
                    middleMan = undefined;
                }
            };
        });
    };
}

/**
 * Delays the emission of items from the source Observable using the microtask queue.
 */
function delayOnMicrotaskQueue() {
    return delay(0, asapScheduler);
}

/**
 * Allows items through whose keys are distinct from the previous item.
 *
 * ```
 * source:                     |-{a:1,b:2}--{a:2,b:3}--{a:2,c:3}--{b:3}--{b:4}-|
 * distinctUntilKeysChanged(): |-{a:1,b:2}-------------{a:2,c:3}--{b:3}--------|
 * ```
 */
function distinctUntilKeysChanged() {
    var lastKeySet;
    return filter(function (value) {
        var keySet = new Set(keys(value));
        if (lastKeySet && isSetEqual(keySet, lastKeySet)) {
            return false;
        }
        lastKeySet = keySet;
        return true;
    });
}

/**
 * Use this to create a complex pipeable operator. It is usually better style to compose existing operators than to create a brand new one, but when you need full control this can reduce some boilerplate.
 *
 * The supplied `subscriber` will act as a simple pass-through of all values, errors, and completion to `destination`. Modify it for your needs.
 *
 * A simple example, recreating the "map" operator:
 * ```ts
 * function map<I, O>(fn: (input: I) => O) {
 *   return createOperatorFunction<I, O>(
 *     (subscriber, destination) => {
 *       subscriber.next = (value) => {
 *         destination.next(fn(value));
 *       };
 *     },
 *   );
 * }
 * ```
 *
 * For a more complex example, check the source of `skipAfter`.
 */
function createOperatorFunction(modifySubscriber) {
    return function (source) {
        return new Observable(function (destination) {
            var subscriber = new Subscriber(destination);
            modifySubscriber(subscriber, destination);
            return source.subscribe(subscriber);
        });
    };
}

/**
 * Works like `filter()`, but always lets through the first emission for each new subscriber. This makes it suitable for subscribers that expect the observable to behave like a `BehaviorSubject`, where the first emission is processed synchronously during the call to `subscribe()` (such as the `async` pipe in an Angular template).
 *
 * ```
 * source:                   |-false--true--false--true--false--true-|
 * filterBehavior(identity): |-false--true---------true---------true-|
 * filterBehavior(identity):        |-true---------true---------true-|
 * filterBehavior(identity):              |-false--true---------true-|
 * ```
 */
function filterBehavior(predicate) {
    return createOperatorFunction(function (subscriber, destination) {
        var firstValue = true;
        subscriber.next = function (value) {
            if (firstValue) {
                destination.next(value);
                firstValue = false;
                return;
            }
            try {
                if (predicate(value)) {
                    destination.next(value);
                }
            }
            catch (ex) {
                destination.error(ex);
            }
        };
    });
}

/**
 * Logs values, errors and completion to the console, and passes them all along unchanged.
 *
 * ```ts
 * of(1, 2).pipe(logValues()).subscribe();
 * // prints using console.log:
 * // [value] 1
 * // [value] 2
 * // [complete]
 *
 * of(1, 2).pipe(logValues("my number", "debug")).subscribe();
 * // prints using console.debug:
 * // [value] my number 1
 * // [value] my number 2
 * // [complete] my number
 *
 * throwError("boo").pipe(logValues("pipe says", "warn")).subscribe();
 * // prints using console.warn:
 * // [error] pipe says boo
 * ```
 */
function logValues(prefix, level) {
    if (level === void 0) { level = "log"; }
    return tap(makeLogFn("[value]"), makeLogFn("[error]"), makeLogFn("[complete]"));
    function makeLogFn() {
        var prefixes = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            prefixes[_i] = arguments[_i];
        }
        if (prefix !== undefined) {
            prefixes.push(prefix);
        }
        return function () {
            var values = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                values[_i] = arguments[_i];
            }
            console[level].apply(console, __spread(prefixes, values));
        };
    }
}

/** @hidden */
function mapAndCacheElements(buildCacheKey, buildDownstreamItem) {
    var cache = new Map();
    return map(function (upstreamItems) {
        var nextCache = new Map();
        var downstreamItems = map$1(upstreamItems, function (upstreamItem, key) {
            var cacheKey = buildCacheKey(upstreamItem, key);
            var downstreamItem;
            if (cache.has(cacheKey)) {
                // tslint:disable-next-line:no-non-null-assertion
                downstreamItem = cache.get(cacheKey);
            }
            else if (nextCache.has(cacheKey)) {
                // tslint:disable-next-line:no-non-null-assertion
                downstreamItem = nextCache.get(cacheKey);
            }
            else {
                downstreamItem = buildDownstreamItem(upstreamItem, key);
            }
            nextCache.set(cacheKey, downstreamItem);
            return downstreamItem;
        });
        cache = nextCache;
        return downstreamItems;
    });
}

/**
 * Applies `buildDownstreamItem` to each item in the upstream array and emits the result. Each downstream item is cached using the key generated by `buildCacheKey` so that the next emission contains references to the matching objects from the previous emission, without running `buildDownstreamItem` again. The cache is only held between successive emissions.
 *
 * This is useful e.g. when using the result in an `*ngFor` expression of an angular template, to prevent angular from rebuilding the inner component and to allow `OnPush` optimizations in the inner component.
 *
 * If multiple items in an upstream array have the same cache key, it will only call `buildDownstreamItem` once.
 *
 * ```ts
 * const mapWithCaching = mapAndCacheArrayElements(
 *   (item) => item,
 *   (item) => item + 1
 * )
 * ```
 * ```
 * source:         -[1, 2]---[1, 2, 3]---[2]--|
 * mapWithCaching: -[2, 3]---[2, 3, 4]---[3]--|
 * ```
 */
var mapAndCacheArrayElements = mapAndCacheElements;

/**
 * Applies `buildDownstreamItem` to each item in the upstream object and emits an array containing the results. Each downstream item is cached using the key generated by `buildCacheKey` so that the next emission contains references to the matching objects from the previous emission, without running `buildDownstreamItem` again. The cache is only held between successive emissions.
 *
 * This is useful e.g. when using the result in an `*ngFor` expression of an angular template, to prevent angular from rebuilding the inner component and to allow `OnPush` optimizations in the inner component.
 *
 * If multiple items in an upstream object have the same cache key, it will only call `buildDownstreamItem` once.
 *
 * ```ts
 * const mapWithCaching = mapAndCacheObjectElements(
 *   (item, key) => key,
 *   (item, key) => item + 1
 * )
 * ```
 * ```
 * source:         -{ a: 1, b: 2 }---{ a: 1, b: 2, c: 3 }---{ b: 2 }--|
 * mapWithCaching: -[2, 3]-----------[2, 3, 4]--------------[3]--|
 * ```
 */
var mapAndCacheObjectElements = mapAndCacheElements;

/**
 * Emits the latest value of the given Observable every time the source Observable emits a value.
 *
 * ```
 * source:               -1---2--3------4-|
 * inner:                ---a------b--c---|
 * mapToLastFrom(inner): -----a--a------c-|
 * ```
 */
function mapToLatestFrom(inner$) {
    return flow(withLatestFrom(inner$), map(function (_a) {
        var _b = __read(_a, 2), _ = _b[0], inner = _b[1];
        return inner;
    }));
}

/**
 * Causes the next value in the pipe to be skipped after `skip$` emits a value. For example:
 *
 * ```
 * source: -1-----2-----3-----4-----5-|
 * skip$:  ----0----------0-0----------
 *
 * result: -1-----------3-----------5-|
 * ```
 * ```ts
 * const source = new Subject();
 * const skip$ = new Subject();
 * const result = source.pipe(skipAfter(skip$));
 *
 * source.next(1); // result emits `1`
 *
 * skip$.next();
 * source.next(2); // result does not emit
 * source.next(3); // result emits `3`
 *
 * skip$.next();
 * skip$.next();
 * source.next(4); // result does not emit
 * source.next(5); // result emits `5`
 * ```
 */
function skipAfter(skip$) {
    return createOperatorFunction(function (subscriber, destination) {
        var skipNext = false;
        subscriber.add(skip$.subscribe(function () {
            skipNext = true;
        }, bindKey(destination, "error")));
        subscriber.next = function (value) {
            if (skipNext) {
                skipNext = false;
            }
            else {
                destination.next(value);
            }
        };
    });
}

/**
 * Emits the upstream value as the first element in an array, followed by the last `count` values in reverse chronological order.
 *
 * ```
 * source:         -1----2------3--------4--------5--------|
 * withHistory(2): -[1]--[2,1]--[3,2,1]--[4,3,2]--[5,4,3]--|
 * withHistory(0): -[1]--[2]----[3]------[4]------[5]------|
 * ```
 */
function withHistory(count) {
    return scan(function (buf, value) { return __spread([value], buf.slice(0, count)); }, []);
}

/**
 * Tracks all subscriptions to easily unsubscribe from them all during cleanup. Also binds callbacks to `this` for convenient use as a superclass, e.g.:
 *
 * ```ts
 * class EventLogger extends SubscriptionManager {
 *   constructor(private prefix: string, event$: Observable<string>) {
 *     super();
 *
 *     // you can pass in an instance method here and it will be bound to `this`
 *     this.subscribeTo(event$, this.log);
 *   }
 *
 *   log(event: string) {
 *     // even though this is used as a callback, you can still use `this`
 *     console.log(this.prefix + event);
 *   }
 * }
 * ```
 */
var SubscriptionManager = /** @class */ (function () {
    function SubscriptionManager() {
        this.subscriptions = new Subscription();
    }
    SubscriptionManager.prototype.subscribeTo = function (observable, next, error, complete) {
        this.subscriptions.add(observable.subscribe(this.bind(next), this.bind(error), this.bind(complete)));
    };
    SubscriptionManager.prototype.unsubscribe = function () {
        this.subscriptions.unsubscribe();
        this.subscriptions = new Subscription();
    };
    SubscriptionManager.prototype.bind = function (fn) {
        return fn && fn.bind(this);
    };
    return SubscriptionManager;
}());

/*
 * Public API Surface of s-rxjs-utils
 */

/**
 * Generated bundle index. Do not edit.
 */

export { SubscriptionManager, cache, createOperatorFunction, delayOnMicrotaskQueue, distinctUntilKeysChanged, filterBehavior, logValues, mapAndCacheArrayElements, mapAndCacheObjectElements, mapToLatestFrom, skipAfter, withHistory };
//# sourceMappingURL=s-rxjs-utils.js.map
